var EventEmitter = require('events').EventEmitter,
  _ = require('lodash'),
  util = require('./util'),
  Schema = require('./schema');

var Model = module.exports = function Model(name, parent, schema, index){
  this._name = name;
  this._parent = parent;
  this._store = parent._store.get(name);

  for (var i=0, len=index.length; i<len; i++){
    var id = index[i];
    index[i] = isNaN(+id) ? id : +id;
  }

  this._index = index;

  if (schema.constructor === Schema){
    this._schema = schema;

    var methods = schema.methods,
      _this = this;

    for (var i in methods){
      this.__proto__[i] = function(){
        return methods[i].apply(_this, arguments);
      };
    }
  } else {
    this._schema = new Schema(schema);
  }

  this.__defineGetter__('length', function(){
    return this._index.length;
  });
};

require('util').inherits(Model, EventEmitter);

Model.prototype._init = function(index){
  return new Model(this._name, this._parent, this._schema, index);
};

Model.prototype._getRaw = function(id){
  return this._store.get(id);
};

Model.prototype._getOne = function(id){
  var data = this._getRaw(id);
  if (data){
    data = this._schema.restore(data);
    data._id = isNaN(+id) ? id : +id;
    return data;
  }
};

Model.prototype.get = function(){
  var args = util.unique(util.flatten(_.toArray(arguments)));

  if (args.length > 1){
    var results = [];
    for (var i=0, len=args.length; i<len; i++){
      results.push(this._getOne(args[i]));
    }
    return results;
  } else {
    return this._getOne(args[0]);
  }
};

Model.prototype.forEach = Model.prototype.each = function(iterator){
  var index = this._index;
  for (var i=0, len=index.length; i<len; i++){
    var id = index[i];
    iterator(this._getOne(id), id);
  }
  return this;
};

Model.prototype.toArray = function(){
  var results = [];
  this.each(function(item){
    results.push(item);
  });
  return results;
};

Model.prototype.count = function(){
  return this._index.length;
};

Model.prototype._insertOne = function(id, obj, callback){
  var data = this._schema.save(obj);
  id = isNaN(+id) ? id : +id;
  this._store.set(id, data);
  var item = this._getOne(id);
  this.emit('insert', item, id);
  callback(item, id);
};

Model.prototype.insert = function(data, callback){
  if (!_.isFunction(callback)) callback = function(){};

  if (Array.isArray(data)){
    var length = data.length,
      id = this._store.get('_primary'),
      ids = [],
      results = [];
    for (var i=0; i<length; i++){
      this._insertOne(id + i, data[i], function(item){
        results.push(item);
      });
      ids.push(id + i);
    }
    this._index = this._index.concat(ids);
    this._store.set('_primary', id + length);
    callback(results);
  } else {
    var id = this._store.get('_primary'),
      result;
    this._insertOne(id, data, function(item){
      result = item;
    });
    this._index.push(id);
    this._store.set('_primary', id + 1);
    callback(result, id);
  }

  return this;
};

Model.prototype._updateOne = function(id, obj){
  var data = this._getRaw(id);
  id = isNaN(+id) ? id : +id;

  for (var i in obj){
    var target = obj[i],
      origin = data[i];
    if (data.hasOwnProperty(i) && _.isObject(target) && !Array.isArray(target)){
      if (Array.isArray(origin)){
        for (var j in target){
          var prop = target[j];
          switch (j){
            case '$push':
              if (Array.isArray(prop)){
                data[i] = origin.concat(prop);
              } else {
                data[i].push(prop);
              }
              break;
            case '$pull':
              if (Array.isArray(prop)){
                data[i] = _.difference(origin, prop);
              } else {
                data[i] = _.without(origin, prop);
              }
              break;
            case '$shift':
              prop = parseInt(prop, 10);
              if (prop > 0){
                data[i] = origin.slice(prop);
              } else {
                data[i] = origin.slice(0, origin.length + prop);
              }
              break;
            case '$pop':
              prop = parseInt(prop, 10);
              if (prop > 0){
                data[i] = origin.slice(0, origin.length - prop);
              } else {
                data[i] = origin.slice(-prop);
              }
              break;
            case '$addToSet':
              if (Array.isArray(prop)){
                for (var k=0, len=prop.length; k<len; k++){
                  var item = prop[k];
                  if (origin.indexOf(item) === -1){
                    data[i].push(item);
                  }
                }
              } else {
                if (origin.indexOf(prop) === -1){
                  data[i].push(prop);
                }
              }
              break;
          }
        }
      } else if (_.isNumber(origin)){
        for (var j in target){
          var prop = target[j];
          switch (j){
            case '$inc':
              data[i] += prop;
              break;
            case '$dec':
              data[i] -= prop;
              break;
          }
        }
      } else {
        data[i] = target;
      }
    } else {
      data[i] = target;
    }
  }

  this._store.set(id, this._schema.save(data));
  this.emit('update', data, id);
};

Model.prototype.update = function(id, data){
  if (_.isObject(id)){
    var index = this._index;
    for (var i=0, len=index.length; i<len; i++){
      this._updateOne(index[i], id);
    }
  } else {
    this._updateOne(id, data);
  }
  return this;
};

Model.prototype._replaceOne = function(id, data){
  id = isNaN(+id) ? id : +id;
  this._store.set(id, this._schema.save(data));
  this.emit('update', data, id);
};

Model.prototype.replace = function(id, data){
  if (_.isObject(id)){
    var index = this._index;
    for (var i=0, len=index.length; i<len; i++){
      this._replaceOne(index[i], id);
    }
  } else {
    this._replaceOne(id, data);
  }
  return this;
};

Model.prototype._removeOne = function(id){
  id = isNaN(+id) ? id : +id;
  this._store.remove(id);
  this.emit('remove', id);
};

Model.prototype.remove = function(){
  var args = util.unique(util.flatten(_.toArray(arguments))),
    index = this._index;

  if (!args.length){
    for (var i=0, len=index.length; i<len; i++){
      this._removeOne(index[i]);
    }
    this._index = [];
  } else if (args.length > 1){
    for (var i=0, len=args.length; i<len; i++){
      var id = args[i];
      id = args[i] = isNaN(+id) ? id : +id;
      this._removeOne(id);
    }
    this._index = _.difference(index, args);
  } else {
    var id = args[0];
    id = isNaN(+id) ? id : +id;
    this._removeOne(id);
    this._index = _.without(index, id);
  }
  return this;
};

Model.prototype.destroy = function(){
  this._parent._store.remove(this._name);
  return this;
};

Model.prototype.first = function(){
  var id = this._index[0];
  if (id) return this._getOne(id);
};

Model.prototype.last = function(){
  var index = this._index,
    id = index[index.length - 1];
  if (id) return this._getOne(id);
};

Model.prototype.eq = function(i){
  var index = this._index,
    id = i >= 0 ? index[i] : index[index.length + i];
  if (id) return this._getOne(id);
};

Model.prototype.slice = function(start, end){
  return this._init([].slice.apply(this._index, arguments));
};

Model.prototype.limit = function(num){
  return this.slice(0, num);
};

Model.prototype.skip = function(num){
  return this.slice(num);
};

Model.prototype.reverse = function(){
  return this._init(util.reverse(this._index));
};

Model.prototype.sort = function(orderby, order){
  var data = {};

  this.each(function(item, id){
    data[id] = item;
  });

  var arr = util.copy(this._index).sort(function(a, b){
    var orderA = data[a][orderby],
      orderB = data[b][orderby];

    if (orderA < orderB){
      return -1;
    } else if (orderA > orderB){
      return 1;
    } else {
      return 0;
    }
  });

  if (order){
    order = order.toString();
    if (order == -1 || order.toLowerCase() == 'desc') arr = arr.reverse();
  }

  return this._init(arr);
};

Model.prototype.random = Model.prototype.shuffle = function(){
  return this._init(util.shuffle(this._index));
};

Model.prototype.find = function(queries){
  var arr = [];

  this.each(function(item, id){
    var match = true;

    for (var i in queries){
      var query = queries[i],
        item = item[i];

      if (_.isRegExp(query)){
        match = !!item.match(query);
      } else if (_.isObject(query)){
        for (var j in query){
          var rule = query[j];

          switch (j){
            case '$lt':
              match = item < rule;
              break;
            case '$lte':
            case '$max':
              match = item <= rule;
              break;
            case '$gt':
              match = item > rule;
              break;
            case '$gte':
            case '$min':
              match = item >= rule;
              break;
            case '$length':
              var length = item.length;
              if (_.isObject(rule)){
                for (var k in rule){
                  var sub = rule[k];

                  switch (k){
                    case '$lt':
                      match = length < sub;
                      break;
                    case '$lte':
                      match = length <= sub;
                      break;
                    case '$gt':
                      match = length > sub;
                      break;
                    case '$gte':
                      match = length >= sub;
                      break;
                  }

                  if (!match) break;
                }
              } else {
                match = length === rule;
              }
              break;
            case '$in':
              if (!Array.isArray(rule)) rule = [rule];
              var isArr = Array.isArray(item);
              match = false;
              for (var i=0, len=rule.length; i<len; i++){
                if (isArr){
                  if (item.indexOf(rule[i]) !== -1){
                    match = true;
                    break;
                  }
                } else {
                  if (item === rule[i]){
                    match = true;
                    break;
                  }
                }
              }
              break;
            case '$nin':
              if (!Array.isArray(rule)) rule = [rule];
              var isArr = Array.isArray(item);
              for (var i=0, len=rule.length; i<len; i++){
                if (isArr){
                  if (item.indexOf(rule[i]) !== -1){
                    match = false;
                    break;
                  }
                } else {
                  if (item === rule[i]){
                    match = false;
                    break;
                  }
                }
              }
              break;
            case '$all':
              if (!Array.isArray(rule)) rule = [rule];
              for (var i=0, len=rule.length; i<len; i++){
                if (item.indexOf(rule[i]) === -1){
                  match = false;
                  break;
                }
              }
              break;
            case '$exists':
              var exist = item != null && typeof item !== 'undefined';
              match = exist == rule;
              break;
            case '$ne':
              if (_.isObject(item) && _.isObject(rule)){
                match = JSON.stringify(rule) !== JSON.stringify(item);
              } else {
                match = rule !== item;
              }
              break;
          }

          if (!match) break;
        }
      } else {
        if (_.isObject(item) && _.isObject(query)){
          match = JSON.stringify(query) === JSON.stringify(item);
        } else {
          match = query === item;
        }
      }

      if (!match) break;
    }

    if (match) arr.push(id);
  });

  return this._init(arr);
};

Model.prototype.findOne = function(query){
  return this.find(query).first();
};